"""USB helper functions to interface with a Free-Wili."""

import pathlib
import sys
from dataclasses import dataclass

import usb

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import platform

if platform.system().lower() == "windows":
    import warnings

    warnings.filterwarnings("ignore")
    import usb.backend.libusb1
    import usb1
    from pyusb_chain.usb_tree_view_tool import UsbTreeViewTool

# FreeWili USB Hub Vendor ID.
USB_VID_FW_HUB: int = 0x0424
# FreeWili USB Hub Product ID.
USB_PID_FW_HUB: int = 0x2513

# FreeWili Black FTDI VendorID
USB_VID_FW_FTDI: int = 0x0403
# FreeWili Black FTDI ProductID
USB_PID_FW_FTDI: int = 0x6014

# Raspberry Pi Vendor ID
USB_VID_FW_RPI: int = 0x2E8A
# Raspberry Pi Pico SDK CDC UART Product ID
USB_PID_FW_RPI_CDC_PID: int = 0x000A
# Raspberry Pi Pico SDK UF2 Product ID
USB_PID_FW_RPI_UF2_PID: int = 0x0003


@dataclass(frozen=True)
class USBLocationInfo:
    """USB Location Information for a specific device."""

    # USB Vendor ID
    vendor_id: int
    # USB Product ID
    product_id: int
    # USB bus location
    bus: int
    # USB address location
    address: int
    # USB port number
    port_number: int
    # USB port numbers
    port_numbers: int
    # USB product name
    name: None | str
    # USB serial number
    serial: str
    # Parent USB device this is attached to (HUB)
    parent: None | Self

    def __str__(self) -> str:
        return (
            f"{self.name} {self.serial} "
            f"[{self.vendor_id:#04x}:{self.product_id:#04x}]"
            f"[{self.bus}:{self.address}:{self.port_numbers}:{self.port_number}]"
        )

    @classmethod
    def from_libusb(cls, dev: usb.core.Device) -> Self:
        """Create a USBLocationInfo based on libusb object."""
        # Get the serial number
        try:
            dev._langids = (1033,)
            serial_number: str = dev.serial_number
        except Exception as _ex:
            serial_number = ""
        # Get the product name
        try:
            product: str = dev.product
        except Exception as _ex:
            product = ""
        # Try to query the parent device
        try:
            parent = None
            if dev.parent:
                parent = cls.from_libusb(dev.parent)
        except Exception as ex:
            print(f"Exception: from_lubusb(): {ex}")
            parent = None
        # Create our USBLocationInfo
        usb_location_info = cls(
            dev.idVendor,
            dev.idProduct,
            dev.bus,
            dev.address,
            dev.port_number,
            dev.port_numbers,
            product,
            serial_number,
            parent,
        )
        return usb_location_info


def find_all(vid: None | int = None, pid: None | int = None, no_match_raises: bool = True) -> tuple[USBLocationInfo, ...]:
    """Find all USB devices attached to Host from a given USB VID and PID.

    Parameters:
    ----------
        vid: None | int
            USB Vendor ID to search for
        pid: None | int
            USB Product ID to search for
        no_match_raises: bool
            If we can't find a match on windows to USBView, raise an exception.
            This is needed for matching serial numbers and the name.

    Returns:
    -------
        tuple[USBLocationInfo]:
            Tuple of USBLocationInfo

    Raises:
    -------
        None
    """
    # We can't pass idVendor and idProduct in as None because it won't
    # match anything, so lets not include it if its None
    kwargs = {}
    if vid:
        kwargs["idVendor"] = vid
    if pid:
        kwargs["idProduct"] = pid
    if platform.system().lower() == "windows":
        usb1_location = str(pathlib.Path(usb1.__file__).parent / "libusb-1.0.dll")
        backend = usb.backend.libusb1.get_backend(find_library=lambda x: usb1_location)
        try:
            usb_devices = list(usb.core.find(**kwargs, find_all=True, backend=backend))
        except usb.core.NoBackendError as ex:
            print(f"DEBUG: usb1_location: {usb1_location}")
            raise usb.core.NoBackendError(f"Failed to load libusb, is it installed? {ex}") from ex
        # libusb on windows doesn't support the serial so we are going to grab it from usbview
        # and directly modify the libusb class.
        UsbTreeViewTool()
        tool = UsbTreeViewTool()
        tool.scan()
        win_devices = tool.filter(None)
        # ['1-1', '', 'USB Composite Device - 2× Camera', None, 11]
        # ['2-1-1', 'COM6', 'USB Composite Device - COM6', 'E463A8574B103A35', 19]
        # ['2-1-2', 'COM5', 'USB Composite Device - COM5', 'E463A8574B423835', 20]
        # ['2-1-3', 'COM7', 'USB Serial Converter - COM7', 'FW5275', 21]
        # ['3-1', '', 'USB Composite Device - Mouse, Keyboard, 2× HID', '127C395F3038', 12]
        # ['3-2-1', 'COM3', 'USB Composite Device - COM3', 'E463A8574B133839', 14]
        # ['3-2-2', 'COM4', 'USB Composite Device - COM4', 'E463A8574B5E3839', 15]
        # ['3-2-3', 'COM8', 'USB Serial Converter - COM8', 'FW5491', 22]
        # ['3-3-1', '', 'Qualcomm FastConnect 6900 Bluetooth Adapter', None, 0]
        # ['3-5', '', 'Goodix MOC Fingerprint', None, 0]
        for usb_device in usb_devices:
            # for some reason the port_number seems wrong but the bus reflects
            # the correct number here
            matched = False
            for device in win_devices:
                data = device.export_data(True)[0]
                loc_id = data[0]
                try:
                    win_port_numbers = tuple([int(x.split(":")[0]) for x in loc_id.split("-")])
                except ValueError:
                    # Some devices are reporting strange values like 7:Microphone here...
                    continue
                _com_port_name = data[1]
                name = data[2]
                serial = data[3]
                _address = data[4]
                usb_pn_len = len(usb_device.port_numbers)
                win_pn_len = len(win_port_numbers)
                len_diff = win_pn_len - usb_pn_len
                if win_port_numbers[len_diff:] == usb_device.port_numbers:
                    usb_device._serial_number = serial
                    usb_device._product = name
                    matched = True
                    break
            if not matched and no_match_raises:
                raise RuntimeError("Failed to match device.")
    else:
        usb_devices = list(usb.core.find(**kwargs, find_all=True))
    devices: list[USBLocationInfo] = []
    for dev in usb_devices:
        devices.append(USBLocationInfo.from_libusb(dev))
    return tuple(devices)


if __name__ == "__main__":
    ftdi_devices = find_all(vid=USB_VID_FW_FTDI, pid=USB_PID_FW_FTDI)
    for dev in ftdi_devices:
        print(dev)
    fw_devices = find_all(vid=USB_VID_FW_RPI)
    for dev in fw_devices:
        print(dev)

# def main():
#     import usb1
#     with usb1.USBContext() as context:
#         for device in context.getDeviceIterator(skip_on_error=True):
#             # if device.getVendorID() != USB_VID_FW_HUB or device.getProductID() != USB_PID_FW_HUB:
#             #     continue
#             # print(device.getNumConfigurations())
#             if (device.getBusNumber() != 2):
#                 continue
#             print('ID %04x:%04x' % (device.getVendorID(), device.getProductID()), '->'.join(str(x) for x in ['Bus %03i' % (device.getBusNumber(), )] + device.getPortNumberList()), 'Device', device.getDeviceAddress())


# if __name__ == '__main__':
#     main()



"""Module for serial communication with FreeWili boards.

This module provides functionality to find and control FreeWili boards.
"""

import dataclasses
import enum
import functools
import pathlib
import re
import sys
import time
from typing import Any, Callable, Optional

from freewili.framing import ResponseFrame
from freewili.usb_util import USB_PID_FW_FTDI, USB_VID_FW_FTDI

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

import serial
import serial.tools.list_ports
from result import Err, Ok, Result

from freewili.types import FreeWiliProcessorType


class IOMenuCommand(enum.StrEnum):
    """Free-Wili IO menu representation."""

    High = "s"
    Low = "l"
    Toggle = "t"
    Pwm = "p"
    Stream = "o"
    Get = "u"

    @classmethod
    def from_string(cls, value: str) -> Self:
        """Convert a string value to an IOMenuCommand."""
        match value.lower():
            case "high":
                return cls(cls.High)
            case "low":
                return cls(cls.Low)
            case "toggle":
                return cls(cls.Toggle)
            case "pwm":
                return cls(cls.Pwm)
        raise ValueError(f"'{value}' is not a value IOMenuCommand")


@dataclasses.dataclass
class FreeWiliAppInfo:
    """Information of the FreeWili application."""

    processor_type: FreeWiliProcessorType
    version: int

    def __str__(self) -> str:
        desc = f"{self.processor_type.name}"
        if self.processor_type in (FreeWiliProcessorType.Main, FreeWiliProcessorType.Display):
            desc += f" v{self.version}"
        return desc


# Disable menu Ctrl+b
CMD_DISABLE_MENU = b"\x02"
# Enable menu Ctrl+c
CMD_ENABLE_MENU = b"\x03"

# Raspberry Pi Vendor ID
RPI_VID = 0x2E8A
# Raspberry Pi Pico SDK CDC UART Product ID
RPI_CDC_PID = 0x000A
# Raspberry Pi Pico SDK UF2 Product ID
RPI_UF2_PID = 0x0003


@dataclasses.dataclass(frozen=False)
class FreeWiliSerialInfo:
    """Information of the COM Port of the FreeWili."""

    # COM port of the FreeWili, e.g. 'COM1' or '/dev/ttyACM0'
    port: str
    # Serial number. This is the RPi serial number on Main and Display
    serial: str
    # Application information of the FreeWili firmware
    app_info: FreeWiliAppInfo
    # USB Location of the FreeWili, Optional
    location: Optional[str] = None
    # Vendor ID of the FreeWili (0x2E8A), Optional
    vid: Optional[int] = None
    # Product ID of the FreeWili (0x000A), Optional
    pid: Optional[int] = None
    # FreeWili Serial
    fw_serial: None | str = None

    def __str__(self) -> str:
        desc = f"{self.app_info.processor_type.name} ({self.serial} @ {self.port})"
        if self.app_info.processor_type in (FreeWiliProcessorType.Main, FreeWiliProcessorType.Display):
            desc = f"{self.app_info} ({self.serial} @ {self.port})"
        return desc


class FreeWiliSerial:
    """Class representing a serial connection to a FreeWili."""

    # The default number of bytes to write/read at a time
    DEFAULT_SEGMENT_SIZE: int = 8

    def __init__(self, info: FreeWiliSerialInfo, stay_open: bool = False) -> None:
        self._info: FreeWiliSerialInfo = info
        self._serial: serial.Serial = serial.Serial(None, timeout=1.0, exclusive=True)
        # Initialize to disable menus
        self._stay_open: bool = stay_open

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self._info}>"

    def __str__(self) -> str:
        return f"{self._info}"

    @property
    def info(self) -> FreeWiliSerialInfo:
        """Information of the COM Port of the FreeWili."""
        return self._info

    @property
    def stay_open(self) -> bool:
        """Keep serial port open, if True.

        Returns:
            bool
        """
        return self._stay_open

    @stay_open.setter
    def stay_open(self, value: bool) -> None:
        self._stay_open = value

    def close(self, restore_menu: bool = True) -> None:
        """Close the serial port. Use in conjunction with stay_open."""
        if self._serial.is_open:
            if restore_menu:
                self._set_menu_enabled(True)
            self._serial.close()

    @staticmethod
    def needs_open(enable_menu: bool = False, restore_menu: bool = True) -> Callable:
        """Decorator to open and close serial port.

        Expects the class to have an attribute '_serial' that is a serial.Serial object
        and a method '_init_if_necessary' that initializes the serial port.

        Parameters:
        ----------
            enable_menu: bool
                Enable menu if True. Defaults to False.

            restore_menu: bool
                Restore the menu after we are done. Defaults to True.

        Example:
        -------
        >>> class MyClass:
        >>>     @needs_open(False)
        >>>     def my_method(self):
        >>>         pass
        >>>

        """

        def decorator(func: Callable) -> Callable:
            """Decorator function that wraps the given function."""

            @functools.wraps(func)
            def wrapper(self: Self, *args: Optional[Any], **kwargs: Optional[Any]) -> Any | None:
                if not self._serial.is_open:
                    self._serial.port = self._info.port
                    self._serial.open()
                    self._set_menu_enabled(enable_menu)
                try:
                    result = func(self, *args, **kwargs)
                    self._set_menu_enabled(True)
                    return result
                finally:
                    if not self.stay_open:
                        self.close(restore_menu)
                    result = None

            return wrapper

        return decorator

    def __enter__(self) -> Self:
        if not self._serial.is_open:
            self._serial.port = self._info.port
            self._serial.open()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self._serial.is_open:
            self._serial.close()

    def _set_menu_enabled(self, enabled: bool) -> None:
        """Enable or disable menus.

        Parameters:
        ----------
            enabled: bool
                True to enable menus, False to disable.

        Returns:
        -------
            None
        """
        self._serial.reset_output_buffer()
        self._serial.reset_input_buffer()
        cmd = CMD_ENABLE_MENU if enabled else CMD_DISABLE_MENU
        cmd += "\r\n".encode("ascii")
        self._write_serial(cmd)
        self._serial.flush()

        # Wait for menu to be enabled and receive some data
        timeout_sec: float = 2.0
        if enabled:
            start = time.time()
            while time.time() - start <= timeout_sec and self._serial.in_waiting <= 0:
                time.sleep(0.001)
            time.sleep(0.1)

    def _write_serial(self, data: bytes, timeout_sec: float = 0.0) -> Result[str, str]:
        """Write data to the serial port."""
        # print(f"DEBUG: {repr(data)}")
        try:
            length = self._serial.write(data)
            if length != len(data):
                return Err(f"Only wrote {length} of {len(data)} bytes.")
            self._serial.flush()
            time.sleep(timeout_sec)
        except serial.SerialException as e:
            return Err(f"Failed to write serial data: {str(e)}")
        return Ok(f"Wrote {length} bytes successfully.")

    @needs_open(False)
    def set_io(
        self: Self, io: int, menu_cmd: IOMenuCommand, pwm_freq: None | int = None, pwm_duty: None | int = None
    ) -> Result[ResponseFrame, str]:
        """Set the state of an IO pin to high or low.

        Parameters:
        ----------
            io : int
                The number of the IO pin to set.
            menu_cmd : IOMenuCommand
                Whether to set the pin to high, low, toggle, or pwm.
            pwm_freq: None | int
                PWM frequency in Hertz
            pwm_duty: None | int
                PWM Duty cycle (0-100)

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        # s) High [25]
        # l) Low []
        # t) Toggle
        # p) PWM IO
        # u) Get All IOs (hex)
        match menu_cmd:
            case IOMenuCommand.High:
                cmd = f"o\n{menu_cmd.value}\n{io}\n"
            case IOMenuCommand.Low:
                cmd = f"o\n{menu_cmd.value}\n{io}\n"
            case IOMenuCommand.Toggle:
                cmd = f"o\n{menu_cmd.value}\n{io}\n"
            case IOMenuCommand.Pwm:
                if pwm_freq == -1 or pwm_duty == -1:
                    return Err("pwm_freq and pwm_duty args need to be specified")
                cmd = f"o\n{menu_cmd.value}\n{io} {pwm_freq} {pwm_duty}\n"
            case IOMenuCommand.Toggle:
                cmd = f"o\n{menu_cmd.value}\n{io}\n"
            case _:
                return Err(f"{menu_cmd.name} is not supported.")

        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial(cmd.encode("ascii"), 0.1):
            case Ok(_):
                resp = self._wait_for_response_frame()
                return resp
            case Err(e):
                return Err(e)

    @needs_open(False)
    def set_board_leds(self: Self, io: int, red: int, green: int, blue: int) -> Result[ResponseFrame, str]:
        """Set the GUI RGB LEDs.

        Parameters:
        ----------
            io : int
                The number of the IO pin to set.
            red : int
                Red Color 0-255
            green : int
                Green Color 0-255
            blue : int
                Blue Color 0-255

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # k) GUI Functions
        # s) Set Board LED [25 100 100 100]
        cmd = f"k\ns\n{io} {red} {green} {blue}\n"

        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial(cmd.encode("ascii"), 0.0):
            case Ok(_):
                resp = self._wait_for_response_frame()
                return resp
            case Err(e):
                return Err(e)

    @needs_open(False)
    def get_io(self) -> Result[tuple[int], str]:
        """Get all the IO values.

        Parameters:
        ----------
            None

        Returns:
        -------
            Result[tuple[int], str]:
                Ok(tuple[int]) if the command was sent successfully, Err(str) if not.
        """
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        cmd = f"o\n{IOMenuCommand.Get.value}\n"
        match self._write_serial(cmd.encode("ascii"), 0.1):
            case Ok(_):
                resp = self._wait_for_response_frame()
                if resp.is_err():
                    return resp
                resp = resp.unwrap()
                if not resp.is_ok():
                    return Err(f"Failed to get IO values: {resp.response}")
                all_io_values = int(resp.response, 16)
                values = []
                for i in range(32):
                    io_value = (all_io_values >> i) & 0x1
                    values.append(io_value)
                return Ok(values)
            case Err(e):
                return Err(e)

    def _write_and_read_bytes_cmd(self, command: str, data: bytes, data_segment_size: int) -> Result[bytes, str]:
        """Write and read bytes from a command.

        Parameters:
        ----------
            command : str
                The command to send. Should end with a newline.
            data : bytes
                The data to write.
            data_segment_size : int
                The number of bytes to read/write at a time.

        Returns:
        -------
            Result[bytes, str]:
                Ok(bytes) if the command was sent successfully, Err(str) if not.
        """
        hex_reg = re.compile(r"[A-Fa-f0-9]{1,2}")
        read_bytes = bytearray()
        for i in range(0, len(data), data_segment_size):
            str_hex_data = " ".join(f"{i:02X}" for i in data[i : i + data_segment_size])
            self._serial.write(f"{command}{str_hex_data}\n".encode("ascii"))
            read_data = self._serial.readline().strip()
            # if not read_data:
            #     read_data = self._serial.readline().strip()
            for value in hex_reg.findall(read_data.decode()):
                read_bytes += int(value, 16).to_bytes(1, sys.byteorder)
        return Ok(bytes(read_bytes))

    def _wait_for_response_frame(self, timeout_sec: float = 1.0) -> Result[ResponseFrame, str]:
        """Wait for a response frame after sending a command.

        Parameters:
        ----------
            timeout_sec : float
                Time to wait in seconds before we error out.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the response frame was found, Err(str) if not.
        """
        start = time.time()
        while time.time() - start <= timeout_sec:
            line = self._serial.readline().decode("ascii")
            match ResponseFrame.from_raw(line, False):
                case Ok(rf):
                    return Ok(rf)
                case Err(_msg):
                    time.sleep(0.001)
                    continue
        return Err(f"Failed to read response frame in {timeout_sec} seconds")

    @needs_open(False)
    def read_write_spi_data(self, data: bytes) -> Result[bytes, str]:
        """Read and Write SPI data.

        Parameters:
        ----------
            data : bytes
                The data to write.

        Returns:
        -------
            Result[bytes, str]:
                Ok(bytes) if the command was sent successfully, Err(str) if not.
        """
        return self._write_and_read_bytes_cmd("s\n", data, self.DEFAULT_SEGMENT_SIZE)

    @needs_open(True)
    def write_i2c(self, address: int, register: int, data: bytes) -> Result[ResponseFrame, str]:
        """Write I2C data.

        Parameters:
        ----------
            address : int
                The address to write to.
            register : int
                The register to write to.
            data : bytes
                The data to write.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        data_bytes = " ".join(f"{i:02X}" for i in data)
        match self._write_serial(f"i\nw\n{address:02X} {register:02X} {data_bytes}\n".encode("ascii"), 0.0):
            case Ok(_):
                return self._wait_for_response_frame()
            case Err(e):
                return Err(e)

    @needs_open(True)
    def read_i2c(self, address: int, register: int, data_size: int) -> Result[ResponseFrame, str]:
        """Read I2C data.

        Parameters:
        ----------
            address : int
                The address to write to.
            register : int
                The register to write to.
            data_size : int
                The number of bytes to read.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial(f"i\nr\n{address:02X} {register:02X} {data_size}\n".encode("ascii"), 0.0):
            case Ok(_):
                return self._wait_for_response_frame()
            case Err(e):
                return Err(e)

    @needs_open(True)
    def poll_i2c(self) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Run a script on the FreeWili.

        Arguments:
        ----------
            None

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial("i\np\n".encode("ascii"), 0.1):
            case Ok(_):
                return self._wait_for_response_frame()
            case Err(e):
                return Err(e)

    @needs_open(False)
    def show_gui_image(self, fwi_path: str) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Show a fwi image on the display.

        Arguments:
        ----------
            fwi_path: str
                path to the fwi image

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # k) GUI Functions
        # l) Show FWI Image [pip_boy.fwi]
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial(f"k\nl\n{fwi_path}\n".encode("ascii"), 0.1):
            case Ok(_):
                return self._wait_for_response_frame()
            case Err(e):
                return Err(e)

    @needs_open(False)
    def reset_display(self) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Reset the display back to the main menu.

        Arguments:
        ----------
            None

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # k) GUI Functions
        # t) Reset Display
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial("k\nt\n".encode("ascii"), 0.1):
            case Ok(_):
                return self._wait_for_response_frame()
            case Err(e):
                return Err(e)

    @needs_open(False)
    def show_text_display(self, text: str) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Show text on the display.

        Arguments:
        ----------
            text: str
                text to display on screen.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # k) GUI Functions
        # p) Show Text Display
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial(f"k\np\n{text}\n".encode("ascii"), 0.1):
            case Ok(_):
                return self._wait_for_response_frame()
            case Err(e):
                return Err(e)

    @needs_open(False)
    def read_all_buttons(self) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Read all the buttons.

        Arguments:
        ----------
            None

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # k) GUI Functions
        # u) Read All Buttons
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial("k\nu\n".encode("ascii"), 0.1):
            case Ok(_):
                return self._wait_for_response_frame()
            case Err(e):
                return Err(e)

    @needs_open(False)
    def write_radio(self, data: bytes) -> Result[bytes, str]:
        """Write radio data.

        Parameters:
        ----------
            data : bytes
                The data to write.

        Returns:
        -------
            Result[bytes, str]:
                Ok(bytes) if the command was sent successfully, Err(str) if not.
        """
        return self._write_and_read_bytes_cmd("t\n", data, self.DEFAULT_SEGMENT_SIZE)

    @needs_open(False)
    def read_radio(self, data: bytes) -> Result[bytes, str]:
        """Read radio data.

        Parameters:
        ----------
            data : bytes
                The data to write.

        Returns:
        -------
            Result[bytes, str]:
                Ok(bytes) if the command was sent successfully, Err(str) if not.
        """
        return self._write_and_read_bytes_cmd("k\n", data, self.DEFAULT_SEGMENT_SIZE)

    @needs_open(False)
    def write_uart(self, data: bytes) -> Result[bytes, str]:
        """Write uart data.

        Parameters:
        ----------
            data : bytes
                The data to write.

        Returns:
        -------
            Result[bytes, str]:
                Ok(bytes) if the command was sent successfully, Err(str) if not.
        """
        return self._write_and_read_bytes_cmd("u\n", data, self.DEFAULT_SEGMENT_SIZE)

    @needs_open(False)
    def enable_stream(self, enable: bool) -> None:
        """TODO: Docstring."""
        raise NotImplementedError

    @needs_open(False)
    def run_script(self, file_name: str) -> Result[str, str]:
        """Run a script on the FreeWili.

        Arguments:
        ----------
        file_name: str
            Name of the file in the FreeWili. 8.3 filename limit exists as of V12

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        print(f"Running script '{file_name}' on {self}...")
        match self._write_serial(f"w\n{file_name}\n".encode("ascii")):
            case Ok(_):
                read_bytes = []
                while byte := self._serial.read(1):
                    read_bytes.append(byte.decode())
                return Ok("".join(read_bytes))
            case Err(e):
                return Err(e)

    @needs_open(False)
    def load_fpga_from_file(self, file_name: str) -> Result[str, str]:
        """Load an FGPA from a file on the FreeWili.

        Arguments:
        ----------
        file_name: str
            Name of the file in the FreeWili. 8.3 filename limit exists as of V12

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        match self._write_serial(f"m\n{file_name}\n".encode("ascii")):
            case Ok(_):
                read_bytes = []
                while byte := self._serial.read(1):
                    read_bytes.append(byte.decode())
                return Ok("".join(read_bytes))
            case Err(e):
                return Err(e)

    @needs_open(False)
    def send_file(self, source_file: pathlib.Path, target_name: str) -> Result[str, str]:
        """Send a file to the FreeWili.

        Arguments:
        ----------
        source_file: pathlib.Path
            Path to the file to be sent.
        target_name: str
            Name of the file in the FreeWili.

        Returns:
        -------
            Result[str, str]:
                Returns Ok(str) if the command was sent successfully, Err(str) if not.
        """
        if not isinstance(source_file, pathlib.Path):
            source_file = pathlib.Path(source_file)
        if not source_file.exists():
            return Err(f"{source_file} does not exist.")
        fsize = source_file.stat().st_size
        # generate the checksum
        checksum = 0
        with source_file.open("rb") as f:
            while byte := f.read(1):
                checksum += int.from_bytes(byte, "little")
                if checksum & 0x8000:
                    checksum ^= 2054
                checksum &= 0xFFFFFF
        # send the download command
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()
        match self._write_serial(f"x\nf\n{target_name} {fsize} {checksum}\n".encode("ascii"), 0.1):
            case Ok(_):
                # print(self._serial.read_all())
                print(f"Downloading {source_file} ({fsize} bytes) as {target_name} on {self}")
                with source_file.open("rb") as f:
                    while byte := f.read(1):
                        # print(byte)
                        if self._serial.write(byte) != len(byte):
                            return Err(f"Failed to write {byte.decode()} to {self}")
                        # print(self._serial.read_all())
                        # time.sleep(0.002)
                time.sleep(1)
                return Ok(f"Downloaded {source_file} ({fsize} bytes) as {target_name} to {self}")
            case Err(e):
                return Err(e)

    @needs_open(False)
    def get_file(self, source_file: str) -> Result[bytearray, str]:
        """Get a file from the FreeWili.

        Arguments:
        ----------
        source_file: str
            Name of the file in the FreeWili. 8.3 filename limit exists as of V12

        Returns:
        -------
            Result[bytearray, str]:
                Returns an array of bytes if the command was sent successfully, Err(str) if not.
        """
        # Clear anything in the buffer
        _ = self._serial.read_all()
        match self._write_serial(f"x\nu\n{source_file}\n".encode("ascii")):
            case Ok(_):
                time.sleep(1)
                data = self._serial.read_all()
                return Ok(data)
            case Err(e):
                return Err(e)

    def reset_to_uf2_bootloader(self) -> Result[None, str]:
        """Reset the FreeWili to the uf2 bootloader.

        Returns:
        -------
            Result[None, str]:
                Returns Ok(None) if the command was sent successfully, Err(str) if not.
        """
        original_baudrate = self._serial.baudrate
        try:
            if self._serial.is_open:
                self._serial.close()
            else:
                self._serial.port = self._info.port
            self._serial.baudrate = 1200
            self._serial.open()
            time.sleep(0.1)
            self._serial.close()
            return Ok(None)
        except serial.serialutil.SerialException as ex:
            return Err(str(ex))
        finally:
            self._serial.baudrate = original_baudrate

    def _wait_for_serial_data(self, timeout_sec: float, delay_sec: float = 0.1) -> None:
        """Wait for data to be available on the serial port.

        Parameters:
        ----------
            timeout_sec: float
                The maximum amount of time to wait for data.
            delay_sec: float
                The amount of time to wait after checks for data.

        Returns:
        -------
            None

        Raises:
        -------
            TimeoutError
                If the timeout is reached before data is available.
        """
        start = time.time()
        while self._serial.in_waiting == 0:
            time.sleep(0.001)
            if time.time() - start > timeout_sec:
                raise TimeoutError(f"Timed out waiting for data on {self}")
        time.sleep(delay_sec)

    @needs_open(True)
    def get_app_info(self) -> Result[FreeWiliAppInfo, str]:
        """Detect the processor type of the FreeWili.

        Returns:
        -------
            Result[FreeWiliProcessorType, str]:
                Returns Ok(FreeWiliProcessorType) if the command was sent successfully, Err(str) if not.
        """
        self._wait_for_serial_data(3.0)
        data = self._serial.read_all()
        # proc_type_regex = re.compile(r"(Main|Display) Processor")
        # match = proc_type_regex.search(data.decode())
        # if match is None:
        #     return Ok(FreeWiliProcessorType.Unknown)
        # elif "Main Processor" in match.group():
        #     return Ok(FreeWiliProcessorType.Main)
        # elif "Display Processor" in match.group():
        #     return Ok(FreeWiliProcessorType.Display)
        # else:
        #     return Err("Unknown processor type detected!")
        line = ""
        for line in data.decode().splitlines():
            if "Processor" in line:
                break
        proc_type_regex = re.compile(r"(?:Main|Display)|(?:App version)|(?:\d+)")
        results = proc_type_regex.findall(line)
        if len(results) != 3:
            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Unknown, 0))
        processor = results[0]
        version = results[2]
        if "Main" in processor:
            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Main, int(version)))
        elif "Display" in processor:
            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Display, int(version)))
        else:
            return Ok(FreeWiliAppInfo(FreeWiliProcessorType.Unknown, 0))


def find_all(processor_type: Optional[FreeWiliProcessorType] = None) -> tuple[FreeWiliSerial, ...]:
    """Finds all FreeWili connected to the computer.

    Returns:
    -------
        tuple[FreeWiliComPort]
    """
    # All port attributes:
    #
    # description Pico - Board CDC
    # device /dev/ttyACM0
    # device_path /sys/devices/pci0000:00/0000:00:01.2/0000:02:00.0/usb1/1-2/1-2.1/1-2.1:1.0
    # hwid USB VID:PID=2E8A:000A SER=E463A8574B151439 LOCATION=1-2.1:1.0
    # interface Board CDC
    # location 1-2.1:1.0
    # manufacturer Raspberry Pi
    # name ttyACM0
    # pid 10
    # product Pico
    # read_line <bound method SysFS.read_line of <serial.tools.list_ports_linux.SysFS object at 0x7d2135365e80>>
    # serial_number E463A8574B151439
    # subsystem usb
    # usb_description <bound method ListPortInfo.usb_description of <serial.tools.list_ports_linux.SysFS object at 0x7d2135365e80>> # noqa
    # usb_device_path /sys/devices/pci0000:00/0000:00:01.2/0000:02:00.0/usb1/1-2/1-2.1
    # usb_info <bound method ListPortInfo.usb_info of <serial.tools.list_ports_linux.SysFS object at 0x7d2135365e80>>
    # usb_interface_path /sys/devices/pci0000:00/0000:00:01.2/0000:02:00.0/usb1/1-2/1-2.1/1-2.1:1.0
    # vid 11914
    devices: list[FreeWiliSerial] = []
    for port in serial.tools.list_ports.comports():
        if port.vid == USB_VID_FW_FTDI and port.pid == USB_PID_FW_FTDI:
            devices.append(
                FreeWiliSerial(
                    FreeWiliSerialInfo(
                        port.device,
                        port.serial_number,
                        FreeWiliAppInfo(
                            FreeWiliProcessorType.FTDI,
                            0,
                        ),
                        port.location,
                        port.vid,
                        port.pid,
                    )
                ),
            )
        elif port.vid == RPI_VID and port.pid == RPI_CDC_PID:
            devices.append(
                FreeWiliSerial(
                    FreeWiliSerialInfo(
                        port.device,
                        port.serial_number,
                        FreeWiliAppInfo(
                            FreeWiliProcessorType.Unknown,
                            0,
                        ),
                        port.location,
                        port.vid,
                        port.pid,
                    )
                ),
            )
            try:
                # Update the processor type
                app_info = devices[-1].get_app_info().unwrap()
                devices[-1]._info = FreeWiliSerialInfo(
                    port.device,
                    port.serial_number,
                    app_info,
                    port.location,
                    port.vid,
                    port.pid,
                )
                # Filter by processor type
                if processor_type is not None and app_info.processor_type != processor_type:
                    devices.pop()
            except Exception as ex:
                print(ex)
                pass
    return tuple(devices)



"""Test I2C functionality on a FreeWili."""

import pytest

from freewili import FreeWili
from freewili.framing import ResponseFrameType


class NoI2CHardwareError(Exception):
    """Exception to raise when no I2C hardware was found."""

    pass


class I2CHardwareFoundError(Exception):
    """Exception to raise when I2C hardware was found."""

    pass


@pytest.mark.skipif("len(FreeWili.find_all()) == 0")
@pytest.mark.xfail(raises=I2CHardwareFoundError)
def test_hw_i2c_nothing_attached() -> None:
    """Test i2c on a FreeWili with nothing attached."""
    device = FreeWili.find_all()[0]
    device.stay_open = True
    try:
        # Test Polling
        response_frame = device.poll_i2c().expect("Failed to poll i2c")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\p"
        assert response_frame.timestamp != 0
        # assert response_frame.seq_number == 4
        if response_frame.response != "0":
            raise I2CHardwareFoundError("Poll found I2C devices")
        assert response_frame.response == "0"
        assert response_frame.success == 1

        response_frame = device.write_i2c(0x0, 0x0, bytes([0, 1, 2, 3, 4, 5, 6, 7])).expect("Failed to write i2c")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\w"
        assert response_frame.timestamp != 0
        # assert response_frame.seq_number == 4
        # assert response_frame.response == "Invalid"
        assert response_frame.success == 0

        response_frame = device.read_i2c(0x0, 0x0, 8).expect("Failed to read i2c")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\r"
        assert response_frame.timestamp != 0
        # assert response_frame.seq_number == 4
        # assert response_frame.response == "Invalid"
        assert response_frame.success == 0
    finally:
        device.close()


@pytest.mark.skipif("len(FreeWili.find_all()) == 0")
@pytest.mark.xfail(raises=NoI2CHardwareError)
def test_hw_i2c_sparkfun_9dof_imu_breakout() -> None:
    """Test i2c on a FreeWili with SparkFun 9DoF IMU Breakout - ISM330DHCX, MMC5983MA (Qwiic) attached.

    https://www.sparkfun.com/sparkfun-9dof-imu-breakout-ism330dhcx-mmc5983ma-qwiic.html
    ISM330DHCX I2C Address: 0x6B (Default)
    MMC5983MA Magnetometer I2C Address: 0x30
    """
    device = FreeWili.find_all()[0]
    device.stay_open = True

    try:
        # Test Polling
        response_frame = device.poll_i2c().expect("Failed to poll i2c")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\p"
        assert response_frame.timestamp != 0
        if response_frame.response == "0":
            raise NoI2CHardwareError("Poll found nothing")
        assert response_frame.response == "2 30 6B"
        assert response_frame.success == 1

        # Lets read from ISM330DHCX
        # https://cdn.sparkfun.com/assets/d/4/6/d/f/ism330dhcx_Datasheet.pdf
        response_frame = device.read_i2c(0x6B, 0x02, 1).expect("Failed to read address 0x6B on ISM330DHCX")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\r"
        assert response_frame.timestamp != 0
        assert response_frame.response == "3F "
        assert response_frame.success == 1

        # Lets write to ISM330DHCX
        # 0x14 Reset Master logic and output registers. Must be set to ‘1’ and then set it to ‘0’. Default value: 0
        response_frame = device.write_i2c(0x6B, 0x14, bytes([0b10000000])).expect("Failed to write 0x14 on ISM330DHCX")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\w"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok"
        assert response_frame.success == 1
        # Set to 0 to complete the reset
        response_frame = device.write_i2c(0x6B, 0x14, bytes([0b00000000])).expect("Failed to write 0x14 on ISM330DHCX")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\w"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok"
        assert response_frame.success == 1
        # Verify we reset the register to defaults
        response_frame = device.read_i2c(0x6B, 0x14, 1).expect("Failed to read address 0x6B on ISM330DHCX")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\r"
        assert response_frame.timestamp != 0
        # assert response_frame.seq_number == 4
        assert response_frame.response == "00 "
        assert response_frame.success == 1

        # Lets read from MMC5983MA
        # https://cdn.sparkfun.com/assets/a/b/7/7/2/19921-09102019_MMC5983MA_Datasheet_Rev_A-1635338.pdf
        # Product ID1 0x2F
        response_frame = device.read_i2c(0x30, 0x2F, 1).expect("Failed to read register 0x6B on ISM330DHCX")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"i\r"
        assert response_frame.timestamp != 0
        # assert response_frame.seq_number == 4
        assert response_frame.response == "30 "
        assert response_frame.success == 1
    finally:
        device.close()


if __name__ == "__main__":
    import pytest

    pytest.main(
        args=[
            __file__,
            "--verbose",
        ]
    )



"""Module to create a FreeWili image file."""

import pathlib
import struct

from PIL import Image
from result import Err, Ok, Result


def convert(input_file: pathlib.Path, output_file: pathlib.Path) -> Result[str, str]:
    """Convert a PNG or JPEG image to a FreeWili image file (fwi extension).

    Parameters:
    ----------
    input_file : pathlib.Path:
        Path to the input image file (PNG or JPEG).

    output_file : pathlib.Path:
        Path to the output image file (.fwi).

    Returns:
    -------
        Result[str, str]:
            Ok(str) if the command was sent successfully, Err(str) if not.
    """
    try:
        im = Image.open(input_file)
        # print ("/* Image Width:%d Height:%d */" % (im.size[0], im.size[1]))
    except Exception as ex:
        return Err(f"Fail to open png or jpg file {input_file}: {ex}")

    image_height = im.size[1]
    image_width = im.size[0]

    try:
        outfile = open(output_file, "wb")
    except Exception as ex:
        return Err(f"Can't write the file {output_file}: {ex}")

    # HEADER section of FreeWili image (fwi extension)

    headerid = bytearray(bytes("FW01IMG\0", "ascii"))
    outfile.write(headerid)

    print(headerid)

    flags = struct.pack("<I", 1)  # Flags (1=transparent) unsigned int		iImageFlags;
    outfile.write(flags)

    totalpixelcount = struct.pack("<I", im.size[0] * im.size[1])  #  unsigned int		iImageTotalPixelCount;
    outfile.write(totalpixelcount)

    widthbytes = struct.pack("<h", im.size[0])  # unsigned short	iImageWidth;
    outfile.write(widthbytes)

    heightbytes = struct.pack("<h", im.size[1])  # unsigned short	iImageHeight;
    outfile.write(heightbytes)

    imagetransparentcolor = struct.pack("<h", 0)  # unsigned short	iImageTransparentColor; // always 0
    outfile.write(imagetransparentcolor)

    image_id = struct.pack("<h", 0)  # unsigned short	image_id; // always 0
    outfile.write(image_id)

    # do the acutal pixel bytes

    # for each pixel
    pix = im.load()

    hastransparency = False
    if len(pix[0, 0]) == 4:
        hastransparency = True

    for h in range(image_height):
        for w in range(image_width):
            # alternative conversion method
            # R=pix[w,h][0]>>3
            # G=pix[w,h][1]>>2
            # B=pix[w,h][2]>>3

            r = pix[w, h][0] / 255.0 * 31.0
            g = pix[w, h][1] / 255.0 * 63.0
            b = pix[w, h][2] / 255.0 * 31.0

            if hastransparency:
                if pix[w, h][3] == 0:  # transparent
                    rgb = 0  # transparent color
                else:
                    rgb = (int(r) << 11) | (int(g) << 5) | int(b)
            else:
                rgb = (int(r) << 11) | (int(g) << 5) | int(b)

            # swap
            rgb = ((rgb << 8) & 0xFF00) | ((rgb >> 8) & 0xFF)

            outfile.write(struct.pack("<H", rgb))

    outfile.close()

    return Ok(f"""png or JPG file "{input_file}" converted to FreeWili image file "{output_file}""")



"""Test I2C functionality on a FreeWili."""

import pytest

from freewili import FreeWili
from freewili.framing import ResponseFrameType
from freewili.serial_util import IOMenuCommand


@pytest.mark.skipif("len(FreeWili.find_all()) == 0")
def test_hw_io() -> None:
    """Test IO on a FreeWili."""
    device = FreeWili.find_all()[0]
    device.stay_open = True

    try:
        # Set IO low
        response_frame = device.set_io(25, IOMenuCommand.Low).expect("Failed to set IO low")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"o\l"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok"
        assert response_frame.is_ok()
        # Check to make sure IO is low
        assert device.get_io().expect("Failed to get IO")[25] == 0
        # Set IO High
        response_frame = device.set_io(25, IOMenuCommand.High).expect("Failed to set IO high")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"o\s"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok"
        assert response_frame.is_ok()
        # Check to make sure IO is high
        assert device.get_io().expect("Failed to get IO")[25] == 1
        # Set IO toggle to low
        response_frame = device.set_io(25, IOMenuCommand.Toggle).expect("Failed to set IO high")
        assert response_frame.rf_type == ResponseFrameType.Standard
        assert response_frame.rf_type_data == r"o\t"
        assert response_frame.timestamp != 0
        assert response_frame.response == "Ok"
        assert response_frame.is_ok()
        # Check to make sure IO is low
        assert device.get_io().expect("Failed to get IO")[25] == 0
    finally:
        device.close()


if __name__ == "__main__":
    import pytest

    pytest.main(
        args=[
            __file__,
            "--verbose",
        ]
    )



"""For Interfacing to Free-Wili Devices."""

import pathlib
import platform
import sys
from collections import OrderedDict
from dataclasses import dataclass

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from result import Err, Ok, Result

import freewili
from freewili import usb_util
from freewili.framing import ResponseFrame
from freewili.serial_util import FreeWiliSerial, IOMenuCommand
from freewili.types import FreeWiliProcessorType
from freewili.usb_util import (
    USB_PID_FW_FTDI,
    USB_PID_FW_HUB,
    USB_PID_FW_RPI_UF2_PID,
    USB_VID_FW_FTDI,
    USB_VID_FW_HUB,
    USB_VID_FW_RPI,
    USBLocationInfo,
)

# USB Locations:
# first address = FTDI
FTDI_HUB_LOC_INDEX = 2
# second address = Display
DISPLAY_HUB_LOC_INDEX = 1
# third address = Main
MAIN_HUB_LOC_INDEX = 0

# This maps the actual GPIO exposed on the connector, all others not
# listed here are internal to the processor.
GPIO_MAP = {
    8: "GPIO8/UART1_Tx_OUT",
    9: "GPIO9/UART1_Rx_IN",
    10: "GPIO10/UART1_CTS_IN",
    11: "GPIO11/UART1_RTS_OUT",
    12: "GPIO12/SPI1_Rx_IN",
    13: "GPIO13/SPI1_CS_OUT",
    14: "GPIO14/SPI1_SCLK_OUT",
    15: "GPIO15/SPI1_Tx_OUT",
    16: "GPIO16/I2C0 SDA",
    17: "GPIO17/I2C0 SCL",
    25: "GPIO25/GPIO25_OUT",
    26: "GPIO26/GPIO26_IN",
    27: "GPIO27/GPIO_27_OUT",
}


@dataclass(frozen=True)
class FreeWiliProcessorInfo:
    """Processor USB and Serial Port info of the Free-Wili."""

    processor_type: FreeWiliProcessorType
    usb_info: USBLocationInfo
    serial_info: None | FreeWiliSerial

    def __str__(self) -> str:
        if self.serial_info:
            return f"{self.serial_info}"
        return f"{self.processor_type}: {self.usb_info}"


@dataclass(frozen=True)
class FreeWiliInfo:
    """FreeWili Info."""

    serial_number: str
    processors: tuple[FreeWiliProcessorInfo, ...]


class FreeWili:
    """Free-Wili device used to access FTDI and serial functionality."""

    def __init__(self, info: FreeWiliInfo):
        self.info = info
        self._stay_open = False

    def __str__(self) -> str:
        return f"Free-Wili {self.info.serial_number}"

    def _get_processor(self, processor_type: FreeWiliProcessorType) -> FreeWiliProcessorInfo:
        for processor in self.info.processors:
            if processor.processor_type == processor_type:
                if processor.serial_info:
                    processor.serial_info.stay_open = self.stay_open
                return processor
        raise IndexError(f"Processor {processor_type} not found for {self}")

    @property
    def ftdi(self) -> FreeWiliProcessorInfo:
        """Get FTDI processor."""
        return self._get_processor(FreeWiliProcessorType.FTDI)

    @property
    def main(self) -> FreeWiliProcessorInfo:
        """Get Main processor."""
        return self._get_processor(FreeWiliProcessorType.Main)

    @property
    def display(self) -> FreeWiliProcessorInfo:
        """Get Display processor."""
        return self._get_processor(FreeWiliProcessorType.Display)

    @property
    def stay_open(self) -> bool:
        """Keep serial port open, if True.

        Returns:
            bool
        """
        return self._stay_open

    @stay_open.setter
    def stay_open(self, value: bool) -> None:
        self._stay_open = value

    def close(self, restore_menu: bool = True) -> None:
        """Close the serial port. Use in conjunction with stay_open.

        Arguments:
        ----------
            restore_menu: bool
                Restore the menu functionality before close

        Returns:
        -------
            None
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(FreeWiliProcessorType.Main).serial_info
        if serial_info:
            serial_info.close()
        serial_info = self._get_processor(FreeWiliProcessorType.Display).serial_info
        if serial_info:
            serial_info.close()

    @classmethod
    def find_first(cls) -> Result[Self, str]:
        """Find first Free-Wili device attached to the host.

        Parameters:
        ----------
            None

        Returns:
        -------
            Result[FreeWili, str]:
                Ok(FreeWili) if successful, Err(str) otherwise.

        Raises:
        -------
            None
        """
        try:
            devices = cls.find_all()
            if not devices:
                return Err("No FreeWili devices found!")
            return Ok(devices[0])
        except Exception as ex:
            return Err(str(ex))

    @classmethod
    def find_all(cls) -> tuple[Self, ...]:
        """Find all Free-Wili devices attached to the host.

        Parameters:
        ----------
            None

        Returns:
        -------
            tuple[FreeWili, ...]:
                Tuple of FreeWili devices.

        Raises:
        -------
            None
        """
        hubs = usb_util.find_all(USB_VID_FW_HUB, USB_PID_FW_HUB, False)
        if not hubs:
            # no hubs found
            return ()
        # Find all devices we should have on the hub
        all_usb = usb_util.find_all(USB_VID_FW_FTDI, USB_PID_FW_FTDI) + usb_util.find_all(USB_VID_FW_RPI, None)
        fw_hubs = []
        for hub in hubs:
            usb_devices = OrderedDict({})
            for usb in all_usb:
                if usb.parent != hub:
                    continue
                usb_devices[usb.port_number] = usb
            fw_hubs.append(usb_devices)

        serial_ports = freewili.serial_util.find_all()
        freewilis = []
        for fw_hub in fw_hubs:
            indexes = sorted(fw_hub.keys())
            main_usb: USBLocationInfo = fw_hub[indexes[MAIN_HUB_LOC_INDEX]]
            display_usb: USBLocationInfo = fw_hub[indexes[DISPLAY_HUB_LOC_INDEX]]
            ftdi_usb: USBLocationInfo = fw_hub[indexes[FTDI_HUB_LOC_INDEX]]
            # match up the serial port to the USB device
            ftdi_serial = None
            main_serial = None
            display_serial = None
            for serial_port in serial_ports:
                # Windows likes to append letters to the end of the serial numbers...
                if serial_port.info.serial.startswith(ftdi_usb.serial):
                    ftdi_serial = serial_port
                if main_usb.serial == serial_port.info.serial:
                    serial_port.info.fw_serial = ftdi_usb.serial
                    main_serial = serial_port
                if display_usb.serial == serial_port.info.serial:
                    serial_port.info.fw_serial = ftdi_usb.serial
                    display_serial = serial_port
            # Get main processor based on PID
            main_processor_type = FreeWiliProcessorType.Main
            if main_usb.vendor_id == USB_PID_FW_RPI_UF2_PID:
                main_processor_type = FreeWiliProcessorType.MainUF2
            # Get display processor based on PID
            display_processor_type = FreeWiliProcessorType.Display
            if display_usb.vendor_id == USB_PID_FW_RPI_UF2_PID:
                display_processor_type = FreeWiliProcessorType.DisplayUF2

            processors = (
                FreeWiliProcessorInfo(FreeWiliProcessorType.FTDI, ftdi_usb, ftdi_serial),
                FreeWiliProcessorInfo(main_processor_type, main_usb, main_serial),
                FreeWiliProcessorInfo(display_processor_type, display_usb, display_serial),
            )
            serial: str = ftdi_usb.serial if ftdi_usb.serial else "None"
            freewilis.append(FreeWili(FreeWiliInfo(serial, processors)))
        return tuple(freewilis)  # type: ignore

    def send_file(
        self, source_file: str | pathlib.Path, target_name: None | str, processor: None | FreeWiliProcessorType
    ) -> Result[str, str]:
        """Send a file to the FreeWili.

        Arguments:
        ----------
            source_file: pathlib.Path
                Path to the file to be sent.
            target_name: None | str
                Name of the file in the FreeWili. If None, will be determined automatically based on the filename.
            processor: None | FreeWiliProcessorType
                Processor to upload the file to. If None, will be determined automatically based on the filename.

        Returns:
        -------
            Result[str, str]:
                Returns Ok(str) if the command was sent successfully, Err(str) if not.
        """
        try:
            # Auto assign values that are None
            if not target_name:
                target_name = FileMap.from_fname(str(source_file)).to_path(str(source_file))
            if not processor:
                processor = FileMap.from_fname(str(source_file)).processor
        except ValueError as ex:
            return Err(str(ex))
        assert target_name is not None
        assert processor is not None
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.send_file(source_file, target_name)

    def run_script(
        self, file_name: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[str, str]:
        """Run a script on the FreeWili.

        Arguments:
        ----------
            file_name: str
                Name of the file in the FreeWili. 8.3 filename limit exists as of V12
            processor: FreeWiliProcessorType
                Processor to upload the file to.

        Returns:
        -------
            Result[str, str]:
                Ok(str) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.run_script(file_name)

    def get_io(self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main) -> Result[tuple[int], str]:
        """Get all the IO values.

        Parameters:
        ----------
            processor: FreeWiliProcessorType
                Processor to set IO on.

        Returns:
        -------
            Result[tuple[int], str]:
                Ok(tuple[int]) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.get_io()

    def set_io(
        self: Self,
        io: int,
        menu_cmd: IOMenuCommand,
        pwm_freq: None | int = None,
        pwm_duty: None | int = None,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Main,
    ) -> Result[ResponseFrame, str]:
        """Set the state of an IO pin to high or low.

        Parameters:
        ----------
            io : int
                The number of the IO pin to set.
            menu_cmd : IOMenuCommand
                Whether to set the pin to high, low, toggle, or pwm.
            pwm_freq: None | int
                PWM frequency in Hertz
            pwm_duty: None | int
                PWM Duty cycle (0-100)
            processor: FreeWiliProcessorType
                Processor to set IO on.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.set_io(io, menu_cmd, pwm_freq, pwm_duty)

    def set_board_leds(
        self: Self,
        io: int,
        red: int,
        green: int,
        blue: int,
        processor: FreeWiliProcessorType = FreeWiliProcessorType.Display,
    ) -> Result[ResponseFrame, str]:
        """Set the GUI RGB LEDs.

        Parameters:
        ----------
            io : int
                The number of the IO pin to set.
            red : int
                Red Color 0-255
            green : int
                Green Color 0-255
            blue : int
                Blue Color 0-255
            processor: FreeWiliProcessorType
                Processor to set LEDs on.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # k) GUI Functions
        # s) Set Board LED [25 100 100 100]
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.set_board_leds(io, red, green, blue)

    def read_i2c(
        self, address: int, register: int, data_size: int, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[ResponseFrame, str]:
        """Write I2C data.

        Parameters:
        ----------
            address : int
                The address to write to.
            register : int
                The register to write to.
            data_size : int
                The number of bytes to read.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.read_i2c(address, register, data_size)

    def write_i2c(
        self, address: int, register: int, data: bytes, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main
    ) -> Result[ResponseFrame, str]:
        """Write I2C data.

        Parameters:
        ----------
            address : int
                The address to write to.
            register : int
                The register to write to.
            data : bytes
                The data to write.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.write_i2c(address, register, data)

    def poll_i2c(self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Main) -> Result[ResponseFrame, str]:
        """Write I2C data.

        Parameters:
        ----------
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.poll_i2c()

    def show_gui_image(
        self, fwi_path: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Show a fwi image on the display.

        Arguments:
        ----------
            fwi_path: str
                path to the fwi image
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.show_gui_image(fwi_path)

    def show_text_display(
        self, text: str, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Show text on the display.

        Arguments:
        ----------
            text: str
                text to display on screen.
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.show_text_display(text)

    def read_all_buttons(
        self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Read all the buttons.

        Arguments:
        ----------
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.read_all_buttons()

    def reset_display(
        self, processor: FreeWiliProcessorType = FreeWiliProcessorType.Display
    ) -> Result[ResponseFrame, str]:  # Result[Tuple[int, ...], str]:
        """Reset the display back to the main menu.

        Arguments:
        ----------
            processor: FreeWiliProcessorType
                Processor to use.

        Returns:
        -------
            Result[ResponseFrame, str]:
                Ok(ResponseFrame) if the command was sent successfully, Err(str) if not.
        """
        # Get the FreeWiliSerial and use it
        serial_info = self._get_processor(processor).serial_info
        if not serial_info:
            return Err(f"Serial info not available for {processor}")
        return serial_info.reset_display()


@dataclass(frozen=True)
class FileMap:
    """Map file extension to processor type and location."""

    # file extension type (ie. .fwi)
    extension: str
    # processor the file should live on
    processor: FreeWiliProcessorType
    # directory the file type
    directory: str
    # description of the file type
    description: str

    @classmethod
    def from_ext(cls, ext: str) -> Self:
        """Creates a FileMap from a file extension.

        Parameters:
        ----------
            ext: str
                File extension (ie. ".wasm"). Not case sensitive.

        Returns:
        --------
            FileMap

        Raises:
        -------
            ValueError:
                If the extension isn't known.
        """
        ext = ext.lstrip(".").lower()
        mappings = {
            "wasm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
            "wsm": (FreeWiliProcessorType.Main, "/scripts", "WASM binary"),
            "zio": (FreeWiliProcessorType.Main, "/scripts", "ZoomIO script file"),
            "bit": (FreeWiliProcessorType.Main, "/fpga", "FPGA bit file"),
            "sub": (FreeWiliProcessorType.Display, "/radio", "Radio file"),
            "fwi": (FreeWiliProcessorType.Display, "/images", "Image file"),
            "wav": (FreeWiliProcessorType.Display, "/sounds", "Audio file"),
        }
        if ext not in mappings:
            raise ValueError(f"Extension '{ext}' is not a known FreeWili file type")
        return cls(ext, *mappings[ext])

    @classmethod
    def from_fname(cls, file_name: str) -> Self:
        """Creates a FileMap from a file path.

        Parameters:
        ----------
            file_name: str
                File name (ie. "myfile.wasm"). Not case sensitive. Can contain paths.

        Returns:
        --------
            FileMap

        Raises:
        -------
            ValueError:
                If the extension isn't known.
        """
        fpath = pathlib.Path(file_name)
        return cls.from_ext(fpath.suffix)

    def to_path(self, file_name: str) -> str:
        """Creates a file path from the file_name to upload to the FreeWili.

        Parameters:
        ----------
            file_name: str
                File name (ie. "myfile.wasm"). Not case sensitive. Can contain paths.

        Returns:
        --------
            str
                Full file path intended to be uploaded to a FreeWili

        Raises:
        -------
            ValueError:
                If the extension isn't known.
        """
        fpath = pathlib.Path(file_name)
        fpath_str = str(pathlib.Path(self.directory) / fpath.name)
        if platform.system().lower() == "windows":
            fpath_str = fpath_str.replace("\\", "/")
        return fpath_str


if __name__ == "__main__":
    devices = FreeWili.find_all()
    print(f"Found {len(devices)} Free-Wili(s):")
    for i, dev in enumerate(devices):
        print(f"{i}. {dev}")
        ftdi = dev.ftdi
        main = dev.main
        display = dev.display
        print("\tFTDI:   ", ftdi)  # type: ignore
        print("\tMain:   ", main)  # type: ignore
        print("\tDisplay:", display)  # type: ignore
